#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3					// Use modern global access method and strict wave access
#pragma DefaultTab={3,20,4}			// Set default tab width in Igor Pro 9 and later
#pragma IgorVersion=9.0.0.5			// Fixed length string attributes introduced in 9.0.0.5 (37229), attribute creation order tracking added (37251) 
#pragma version=QLA_version			// should be the same as QLA_version
Constant QLA_version=120			// version 1.2.0, this is written to the _NCProperties attribute of the netCDF header

//___________________________________________________________________________________________________	
//	QLA netCDF-4 Library - Quacks Like A netCDF-4
//	A library of functions for writing netCDF-4 and netCDF-4 classic files using built in HDF5 operations
//___________________________________________________________________________________________________	
//
//	Copyright ©2021
//	Chris Reed (FAAM Airborne Laboratory, Building 146, Central Avenue, Cranfield University, Bedfordshire, MK43 0AL, UK)
//	chris.reed@faam.ac.uk
//___________________________________________________________________________________________________	
//
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//___________________________________________________________________________________________________	

// Define attributes which by CF convention should be the same data type as they are attached to
Static StrConstant respectedAttributes = "_FillValue;missing_value;valid_range;valid_max;valid_min;actual_range;actual_max;actual_min;flag_values;flag_masks;"

// library chunking constants
Static Constant overhang = 0						// allow overhanging chunks
Static Constant chunkMaxBytes = 36864				// maximum total size for a given chunk 
Static Constant contiguousMaxBytes = 1048576		// maximum number of bytes a dataset can have before being automatically chunked 
Static Constant unlimChunkSize = 1					// default chunk size for unlimited dimensions

// options for library error reporting
Static Constant error = 1							// mode of library error handler

// HDF5/netCDF-4 constants, DO NOT MODIFY
Static Constant attMaxBytes = 64000 				// 64kb
Static constant strMaxLen = 255 					// 256 -1 null termination character
Static Constant default1DunlimChunk = 4096			// special case for unlimited dimensions if chunkMaxBytes is set very high
Static Constant padValue = 0x00						// padding character for CHAR datasets
Static StrConstant HDF5markup = "DIMENSION_LIST;DIMENSION_SCALE;CLASS;NAME;REFERENCE_LIST;"
Static StrConstant NCmarkup = "_nc3_strict;_NCProperties;_Netcdf4Coordinates;_Netcdf4Dimid;"

//___________________________________________________________________________________________________	
// create a netCDF-4 file, pass 1 for classic to create netCDF-4 classic model, overwrite an existing file with clobber
Function QLA_create(String pathStr,String fNameStr,int classic,int clobber)
	
	double fileID,err ; String errorStr
	
	// check the status of the output path
	PathInfo /S $pathStr
	if (V_flag == 0)
		sprintf errorStr,"There is no symbolic path with the name '%s'",pathStr
		getError("QLA_create",errorStr,3,fileID)
		return 0
	endif
	
	// create the file
	if (clobber)
		HDF5CreateFile /O /Z /P=$pathStr fileID as fNameStr+".nc"
		sprintf errorStr,"'%s.nc' (id: %d) created in %s\r",fNameStr,fileID,S_path
		getError("QLA_create",errorStr,0,fileID)
	else
		HDF5CreateFile /Z /P=$pathStr fileID as fNameStr+".nc"
		if (V_Flag != 0)
			sprintf errorStr,"The netCDF-4 file '%s.nc' already exists; to overwrite set clobber\r",fNameStr
			getError("QLA_create",errorStr,3,fileID)
			err = GetRTError(1) ; return 0 // clear the error state	
		endif
	endif
	
	// write the _NCProperties attribute if the file was created successfully
	if (V_flag == 0)
		QLA_put_glob_att_text("/","_NCProperties",NCProperties(),fileID)
		
		// write the _nc3_strict attribute if classic was set
		if (classic)
			if(setClassic(fileID) != 0)
				sprintf errorStr,"There was an error when setting the file '%s.nc' to classic model\r",fNameStr 
				getError("QLA_create",errorStr,3,fileID)
			endif
		endif
	else // handle other errors from HDF5CreateFile
		sprintf errorStr,"There was a problem creating the netCDF-4 file '%s.nc'\r",fNameStr
		getError("QLA_create",errorStr,3,fileID)
		err = GetRTError(1) ; return 0 // clear the error state	
	endif
	
	return fileID
End
//___________________________________________________________________________________________________	
// create a group in a netCDF-4 file 
Function QLA_def_grp(String path,double fileID)
	
	double groupID ; String errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// create a group if not classic model
	if (!isClassic(fileID))
		HDF5CreateGroup /Z fileID,path,groupID
		if (V_flag != 0)
			sprintf errorStr,"There was a problem creating the netCDF-4 group '%s'\r",ParseFilePath(0,path,"/",1,0)
			getError("QLA_def_grp",errorStr,3,fileID)
			return 0
		else 
			sprintf errorStr,"Group '%s' created\r",ParseFilePath(0,path,"/",1,0)
			getError("QLA_def_grp",errorStr,0,fileID)
		endif
	endif
	
	return groupID
End
//___________________________________________________________________________________________________	
// write a netCDF-4 dimension variable
Function QLA_def_dim_var(Wave Wv,String path,int type,int deflate,int shuffle,double fileID [,int storage,Wave chunksizes])
	
	int Wvtype=WaveType(Wv),dims=WaveDims(Wv),l,valid,chunk; String WvStr = ParseFilePath(0,path,"/",1,0),error,errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check the name for CF compliance
	if (!isCFName(WvStr))
		sprintf errorStr,"The Variable name '%s' is allowed in netCDF-4 but is not a CF compliant name\r",WvStr
		getError("QLA_def_dim_var",errorStr,2,fileID)
	endif
	
	// dimemsions can only be 1D unless char
	if ((dims > 1) && (type != 1))
		sprintf errorStr,"Dimensions (co-ordinates) can only be 1-Dimensional - '%s' has %d\r",WvStr,dims
		getError("QLA_def_dim_var",errorStr,3,fileID)
		return -1 
	endif
	
	// dimensions should be monotonic and not have invalid values in them
	[valid,error] = isValidCoord(Wv)
	if (!valid)
		sprintf errorStr,"'%s' dimension (co-ordinate) %s\r",WvStr,error
		getError("QLA_def_dim_var",errorStr,2,fileID)
	endif
	
	// create local copy of data in case it must be redimensioned
	Duplicate /O /FREE Wv outWv
	
	// change the wavetype if requested
	if ((type > 1) && (Wvtype != type))
		Redimension /Y=(type) outWv
		Wvtype = type
	endif 
	
	// check if this netCDF is classic model
	if (isClassic(fileID))
		path = WvStr // make sure the path refers to root
		
		// make sure the variable name isn't too long
		if (strlen(WvStr) > strMaxLen)
			sprintf errorStr,"The name '%s' exceeds the %d byte limit for variable names in netCDF-4 classic model and will be clipped\r",strMaxLen,Wvstr 
			getError("QLA_def_dim_var",errorStr,2,fileID)
			WvStr = CleanUpName(WvStr,1,strMaxLen-1)
		endif
		
		// warn for illegal data types, commute if necessary
		if (Wvtype > 32)
			type = isClassicDataType(Wvtype)
			sprintf errorStr,"'%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",WvStr,waveType2String(type)
			getError("QLA_def_dim_var",errorStr,1,fileID)
			Redimension /Y=(type) outWv
		elseif (Wvtype == 0)
			sprintf errorStr,"'%s' is of string or char data type and is not a valid type for dimensions\r",WvStr
			getError("QLA_def_dim_var",errorStr,3,fileID)
			return 0
		endif
	endif

	// is the name of the coordinate the same as one of the existing variables?  
	// there is no way fix this now so bail out
	if (sameName(Path,fileID))
		sprintf errorStr,"'%s' already exists as a variable\r",WvStr
		getError("QLA_def_dim_var",errorStr,1,fileID)
		return -1 
	endif
	
	// get some chunksizes
	Make /O /FREE /N=4 dimsWv
	dimsWv[0] = numpnts(Wv)
	Wave chunkWv = getChunkSizes(outWv,dimsWv)
	
	// does the data need chunking
	chunk = (WaveMin(dimsWv) == -1) || (deflate > 0) || (shuffle == 1) || (storage == 1) || (ParamIsDefault(storage) && (NumberByKey("SIZEINBYTES",WaveInfo(outWv,0)) > contiguousMaxBytes)) ? 1 : 0
		
	if (!ParamIsDefault(chunksizes) && WaveExists(chunksizes))
		l = numpnts(chunksizes)
		if (l == dims)
			chunkWv[0,l-1] = chunksizes[p]
		elseif (chunk)
			sprintf errorStr,"The number of chunks specified (%d) does not match the number of dimensions (%d) of '%s'\r",l,dims,WvStr
			getError("QLA_def_dim_var",errorStr,2,fileID)
		endif
	endif
	
	// write the dimension variable
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /LAYO={2^chunk,chunkWv[0],chunkWv[1],chunkWv[2],chunkWv[3]} /GZIP={deflate,shuffle} /MAXD={dimsWv[0],dimsWv[1],dimsWv[2],dimsWv[3]} outWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the dimension variable '%s'\r",WvStr
		getError("QLA_def_dim_var",errorStr,3,fileID)
		return -1
	endif
	
	// attach the dimension scale to it
	HDF5DimensionScale /Z dataset={fileID,path}, dimName=WvStr, setScale 
	
	// handle errors from HDF5DimesnionScale
	if (V_flag != 0)
		sprintf errorStr,"There was a problem setting the dimension for co-ordinate '%s'\r",WvStr 
		getError("QLA_def_dim_var",errorStr,3,fileID)
		return -1
	endif
	
	// set the _Netcdf4Dimid attribute correctly 
	setDimID(path,getNextNCdimID(fileID),fileID)
	
	// get an error string to display 
	sprintf errorStr,"'%s' co-ordinate variable created\r",WvStr 
	getError("QLA_def_dim_var",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End 
//___________________________________________________________________________________________________	
// write a netCDF-4 dimension - hidden variable
Function QLA_def_dim(Wave Wv,String path,int type,int deflate,int shuffle,double fileID [,int storage,Wave chunksizes])
	
	int WvType=WaveType(Wv),dims=WaveDims(Wv),l,valid,chunk ; String WvStr = ParseFilePath(0,path,"/",1,0),error,errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check the name for CF compliance
	if (!isCFName(WvStr))
		sprintf errorStr,"The Variable name '%s' is allowed in netCDF-4 but is not a CF compliant name\r",WvStr
		getError("QLA_def_dim",errorStr,2,fileID)
	endif

	// dimensions can only be 1D unless char
	if ((dims > 1) && (type != 1))
		sprintf errorStr,"Dimensions (co-ordinates) can only be 1-Dimensional - '%s' has %d\r",WvStr,dims
		getError("QLA_def_dim",errorStr,3,fileID)
		return -1 
	endif
	
	// dimensions should be monotonic and not have invalid values in them
	[valid,error] = isValidCoord(Wv)
	if (!valid)
		sprintf errorStr,"'%s' dimension (co-ordinate) %s\r",WvStr,error
		getError("QLA_def_dim",errorStr,1,fileID)
	endif
	
	// create local copy of data in case it must be redimensioned
	Duplicate /O /FREE Wv outWv
	
	// change the wavetype if requested
	if ((type > 1) && (Wvtype != type))
		Redimension /Y=(type) outWv
		Wvtype = type
	endif 
	
	// check if this netCDF is classic model
	if (isClassic(fileID))
		path = WvStr // make sure the path refers to root
		
		// make sure the variable name isn't too long
		if (strlen(WvStr) > strMaxLen)
			sprintf errorStr,"The name '%s' exceeds the %d byte limit for variable names in netCDF-4 classic model and will be clipped\r",strMaxLen,Wvstr 
			getError("QLA_def_dim",errorStr,2,fileID)
			WvStr = CleanUpName(WvStr,1,strMaxLen-1)
		endif
		
		// warn for illegal data types, commute if necessary
		if (Wvtype > 32)
			type = isClassicDataType(Wvtype)
			sprintf errorStr,"'%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",WvStr,waveType2String(type)
			getError("QLA_def_dim",errorStr,1,fileID)
			Redimension /Y=(type) outWv
		elseif (Wvtype == 0)
			sprintf errorStr,"'%s' is of string or char data type and is not a valid type for dimensions\r",WvStr
			getError("QLA_def_dim",errorStr,3,fileID)
			return 0
		endif
	endif
	
	// is the name of the coordinate the same as one of the existing variables?  
	// there is no way fix this now so bail out
	if (sameName(Path,fileID))
		sprintf errorStr,"'%s' already exists as a variable\r",WvStr
		getError("QLA_def_dim",errorStr,3,fileID)
		return -1 
	endif
	
	// get some chunksizes
	Make /O /FREE /N=4 dimsWv
	dimsWv[0] = numpnts(Wv)
	Wave chunkWv = getChunkSizes(outWv,dimsWv)
	
	// does the data need chunking
	chunk = (WaveMin(dimsWv) == -1) || (deflate > 0) || (shuffle == 1) || (storage == 1) || (ParamIsDefault(storage) && (NumberByKey("SIZEINBYTES",WaveInfo(outWv,0)) > contiguousMaxBytes)) ? 1 : 0
	
	if (!ParamIsDefault(chunksizes))
		l = numpnts(chunksizes)
		if (l == dims)
			chunkWv[0,l-1] = chunksizes[p]
		elseif (chunk)
			sprintf errorStr,"The number of chunks specified (%d) does not match the number of dimensions (%d) of '%s'\r",l,dims,WvStr
			getError("QLA_def_dim",errorStr,2,fileID)
		endif
	endif
	
	// hidden co-ordinate - tell the netCDF API this with the dimName
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /LAYO={2^chunk,chunkWv[0],chunkWv[1],chunkWv[2],chunkWv[3]} /GZIP={deflate,shuffle} /MAXD={dimsWv[0],dimsWv[1],dimsWv[2],dimsWv[3]} outWv, fileID, path

	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the dimension variable '%s'\r",WvStr
		getError("QLA_def_dim",errorStr,3,fileID)
		return -1
	endif
	
	// attach the dimension scale to it
	HDF5DimensionScale /Z dataset={fileID,path}, dimName="This is a netCDF dimension but not a netCDF variable.\t\t"+num2istr(numpnts(Wv)), setScale 
	
	// handle errors from HDF5DimesnionScale
	if (V_flag != 0)
		sprintf errorStr,"There was a problem setting the dimension for co-ordinate '%s'\r",WvStr
		getError("QLA_def_dim",errorStr,3,fileID)
		return -1
	endif
	
	// set the _Netcdf4Dimid attribute correctly 
	setDimID(path,getNextNCdimID(fileID),fileID)
	
	// get an error string to display 
	sprintf errorStr,"'%s' co-ordinate created\r",WvStr 
	getError("QLA_def_dim",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// write an unlimited netCDF-4 dimension  - 1 per netCDF allowed in classic model
Function QLA_def_unlim_dim(Wave Wv,String path,int type,int deflate,int shuffle,double fileID [,Wave chunksizes,int extend])
	
	int WvType=WaveType(Wv),dims=WaveDims(Wv),l,valid ; String WvStr = ParseFilePath(0,path,"/",1,0),error,errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// set the append option correctly 
	extend = (ParamIsDefault(extend)) ? -1 : 0
		
	// check the name for CF compliance
	if (!isCFName(WvStr))
		sprintf errorStr,"The variable name '%s' is allowed in netCDF-4 but is not a CF compliant name\r",WvStr
		getError("QLA_def_unlim_dim",errorStr,2,fileID)
	endif
	
	// dimensions can only be 1D unless char
	if ((dims > 1) && (type != 1))
		sprintf errorStr,"Dimensions (co-ordinates) can only be 1-Dimensional - '%s' has %d\r",WvStr,dims
		getError("QLA_def_unlim_dim",errorStr,3,fileID)
		return -1 
	endif
	
	// dimensions should be monotonic and not have invalid values in them
	[valid,error] = isValidCoord(Wv)
	if (!valid)
		sprintf errorStr,"'%s' dimension (co-ordinate) %s\r",WvStr,error
		getError("QLA_def_unlim_dim",errorStr,1,fileID)
	endif
	
	// create local copy of data in case it must be redimensioned
	Duplicate /O /FREE Wv outWv
	
	// change the wavetype if requested
	if ((type > 1) && (Wvtype != type))
		Redimension /Y=(type) outWv
		Wvtype = type
	endif 
	
	// check if this netCDF is classic model
	if (isClassic(fileID))
		path = WvStr // make sure the path refers to root
		
		// make sure the variable name isn't too long
		if (strlen(WvStr) > strMaxLen)
			sprintf errorStr,"The name '%s' exceeds the %d byte limit for variable names in netCDF-4 classic model and will be clipped\r",strMaxLen,Wvstr 
			getError("QLA_def_unlim_dim",errorStr,2,fileID)
			WvStr=CleanUpName(WvStr,1,strMaxLen-1)
		endif
		
		// warn for illegal data types, commute if necessary
		if (WvType > 32)
			type = isClassicDataType(WvType)
			sprintf errorStr,"'%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",WvStr,waveType2String(type)
			getError("QLA_def_unlim_dim",errorStr,1,fileID)
			Redimension /Y=(type) outWv
		elseif (Wvtype == 0)
			sprintf errorStr,"'%s' is of string or char data type and is not a valid type for dimensions\r",WvStr
			getError("QLA_def_unlim_dim",errorStr,3,fileID)
			return 0
		endif
		
		// check if this is the first unlimited dimension, only 1 is allowed in netCDF-4 classic model
		if (UDimExists(fileID))
			sprintf errorStr,"An unlimited dimension already exists in the output file, '%s' has been created as a standard dimension as only 1 per file is permitted in netCDF-4 classic model\r",WvStr
			getError("QLA_def_unlim_dim",errorStr,2,fileID)
			QLA_def_dim_var(Wv,path,type,deflate,shuffle,fileID,storage=1,chunksizes=chunksizes)
			return 0
		endif
	endif
	
	// is the name of the coordinate the same as one of the existing variables? 
	// there is no way fix this now so bail out
	if (sameName(Path,fileID))
		sprintf errorStr,"'%s' already exists as a variable\r",WvStr
		getError("QLA_def_unlim_dim",errorStr,3,fileID)
		return -1 
	endif
	
	// get some chunksizes
	Make /O /FREE /N=4 dimsWv
	dimsWv[0]=-1
	Wave chunkWv = getChunkSizes(outWv,dimsWv)
	
	if (!ParamIsDefault(chunksizes))
		l = numpnts(chunksizes)
		if (l == dims)
			chunkWv[0,l-1] = chunksizes[p]
		else
			sprintf errorStr,"The number of chunks specified (%d) does not match the number of dimensions (%d) of '%s'\r",l,dims,WvStr
			getError("QLA_def_unlim_dim",errorStr,2,fileID)
		endif
	endif
	
	// write the unlimited dimension variable
	HDF5SaveData /Z /IGOR=0 /APND=(extend) /OPTS=1 /LAYO={2,chunkWv[0],chunkWv[1],chunkWv[2],chunkWv[3]} /GZIP={deflate,shuffle} /MAXD={dimsWv[0],dimsWv[1],dimsWv[2],dimsWv[3]} outWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the dimension variable '%s'\r",WvStr
		getError("QLA_def_unlim_dim",errorStr,3,fileID)
		return -1
	endif
	
	// attach the dimension scale to it
	HDF5DimensionScale /Z dataset={fileID,path}, dimName="This is a netCDF dimension but not a netCDF variable.\t\t"+num2istr(numpnts(Wv)), setScale 
	
	// handle errors from HDF5DimesnionScale
	if (V_flag != 0)
		sprintf errorStr,"There was a problem setting the dimension for co-ordinate '%s'\r",WvStr
		getError("QLA_def_unlim_dim",errorStr,2,fileID)
		return -1
	endif
	
	// set the _Netcdf4Dimid attribute correctly 
	setDimID(path,getNextNCdimID(fileID),fileID)
	
	// get an error string to display 
	sprintf errorStr,"'%s' unlimited co-ordinate created\r",WvStr 
	getError("QLA_def_unlim_dim",errorStr,0,fileID)
		
	HDF5FlushFile fileID
	return 1 
End
//___________________________________________________________________________________________________	
// write an unlimited netCDF-4 dimension variable - 1 per netCDF allowed in classic model
Function QLA_def_unlim_dim_var(Wave Wv,String path,int type,int deflate,int shuffle,double fileID [,Wave chunksizes,int extend])
	
	int WvType=WaveType(Wv),dims=WaveDims(Wv),l,valid ; String WvStr = ParseFilePath(0,path,"/",1,0),error,errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// set the append option correctly 
	extend = (ParamIsDefault(extend)) ? -1 : 0
		
	// check the name for CF compliance
	if (!isCFName(WvStr))
		sprintf errorStr,"The variable name '%s' is allowed in netCDF-4 but is not a CF compliant name\r",WvStr
		getError("QLA_def_unlim_dim_var",errorStr,2,fileID)
	endif
	
	// dimensions can only be 1D unless char
	if ((dims > 1) && (type != 1))
		sprintf errorStr,"Dimensions (co-ordinates) can only be 1-Dimensional - '%s' has %d\r",WvStr,dims
		getError("QLA_def_unlim_dim_var",errorStr,3,fileID)
		return -1 
	endif
	
	// dimensions should be monotonic and not have invalid values in them
	[valid,error] = isValidCoord(Wv)
	if (!valid)
		sprintf errorStr,"'%s' dimension (co-ordinate) %s\r",WvStr,error
		getError("QLA_def_unlim_dim_var",errorStr,2,fileID)
	endif
	
	// create local copy of data in case it must be redimensioned
	Duplicate /O /FREE Wv outWv
	
	// change the wavetype if requested
	if ((type > 1) && (Wvtype != type))
		Redimension /Y=(type) outWv
		Wvtype = type
	endif 
	
	// check if this netCDF is classic model
	if (isClassic(fileID))
		path = WvStr // make sure the path refers to root
		
		// make sure the variable name isn't too long
		if (strlen(WvStr) > strMaxLen)
			sprintf errorStr,"The name '%s' exceeds the %d byte limit for variable names in netCDF-4 classic model and will be clipped\r",strMaxLen,Wvstr 
			getError("QLA_def_unlim_dim_var",errorStr,2,fileID)
			WvStr=CleanUpName(WvStr,1,strMaxLen-1)
		endif
		
		// warn for illegal data types, commute if necessary
		if (WvType > 32)
			type = isClassicDataType(WvType)
			sprintf errorStr,"'%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",WvStr,waveType2String(type)
			getError("QLA_def_unlim_dim_var",errorStr,1,fileID)
			Redimension /Y=(type) outWv
		elseif (Wvtype == 0)
			sprintf errorStr,"'%s' is of string or char data type and is not a valid type for dimensions\r",WvStr
			getError("QLA_def_unlim_dim_var",errorStr,3,fileID)
			return 0
		endif
		
		// check if this is the first unlimited dimension, only 1 is allowed in netCDF-4 classic model
		if (UDimExists(fileID))
			sprintf errorStr,"An unlimited dimension already exists in the output file, '%s' has been created as a standard dimension as only 1 per file is permitted in netCDF-4 classic model\r",WvStr
			getError("QLA_def_unlim_dim_var",errorStr,2,fileID)
			QLA_def_dim_var(Wv,path,type,deflate,shuffle,fileID)
			return 0
		endif
	endif
	
	// is the name of the coordinate the same as one of the existing variables? 
	// there is no way fix this now so bail out
	if (sameName(Path,fileID))
		sprintf errorStr,"'%s' already exists as a variable\r",WvStr
		getError("QLA_def_unlim_dim_var",errorStr,3,fileID)
		return -1 
	endif
	
	// get some chunksizes
	Make /O /FREE /N=4 dimsWv
	dimsWv[0]=-1
	Wave chunkWv = getChunkSizes(outWv,dimsWv)
	
	if (!ParamIsDefault(chunksizes))
		l = numpnts(chunksizes)
		if (l == dims)
			chunkWv[0,l-1] = chunksizes[p]
		else
			sprintf errorStr,"The number of chunks specified (%d) does not match the number of dimensions (%d) of '%s'\r",l,dims,WvStr
			getError("QLA_def_unlim_dim_var",errorStr,2,fileID)
		endif
	endif
	
	// write the unlimited dimension variable
	HDF5SaveData /Z /IGOR=0 /APND=(extend) /OPTS=1 /LAYO={2,chunkWv[0],chunkWv[1],chunkWv[2],chunkWv[3]} /GZIP={deflate,shuffle} /MAXD={dimsWv[0],dimsWv[1],dimsWv[2],dimsWv[3]} outWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the dimension variable '%s'\r",WvStr
		getError("QLA_def_unlim_dim_var",errorStr,3,fileID)
		return -1
	endif
	
	// attach the dimension scale to it
	HDF5DimensionScale /Z dataset={fileID,path}, dimName=WvStr, setScale 
	
	// handle errors from HDF5DimesnionScale
	if (V_flag != 0)
		sprintf errorStr,"There was a problem setting the dimension for co-ordinate '%s'\r",WvStr
		getError("QLA_def_unlim_dim_var",errorStr,3,fileID)
		return -1
	endif
	
	// set the _Netcdf4Dimid attribute correctly 
	setDimID(path,getNextNCdimID(fileID),fileID)
	
	// get an error string to display 
	sprintf errorStr,"'%s' unlimited co-ordinate variable created\r",WvStr 
	getError("QLA_def_unlim_dim_var",errorStr,0,fileID)
	
	
	HDF5FlushFile fileID
	return 1 
End
//___________________________________________________________________________________________________	
// attach netCDF-4 dimensions to a variable 
Function QLA_put_dim(Wave Wv,String path,String dimPathList,double fileID)
	
	int i,j=ItemsInList(dimPathList),dims=WaveDims(Wv),l,classic ; String dimPathStr, WvStr = ParseFilePath(0,path,"/",1,0),scalePath,errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// is the name of the variable the same as one of the coordinates in the same group?
	if (GrepString(dimPathList,path))
		sprintf errorStr,"'%s' already exists as a dimension, cannot attach a dimension  of the same name\r",WvStr
		getError("QLA_put_dim",errorStr,3,fileID)
		return -1
	endif
	
	// check the name for CF compliance
	if (!isCFName(WvStr))
		sprintf errorStr,"The variable name '%s' is allowed in netCDF-4 but is not a CF compliant name\r",WvStr
		getError("QLA_put_dim",errorStr,2,fileID)
	endif
		
	// number of dimensions to attach must equal the number of dimensions
	if (dims != j)
		sprintf errorStr,"Number of dimensions in the dataset '%s' (%d) does not match the number of dimensions to attach (%d)\r",WvStr,dims,j
		getError("QLA_put_dim",errorStr,3,fileID)
		return -1  
	endif	
	
	// check if this netCDF is classic model
	if (isClassic(fileID))
		classic = 1 ; path = WvStr // make sure the path refers to root
		
		// make sure the variable name isn't too long
		if (strlen(WvStr)>strMaxLen)
			sprintf errorStr,"The name '%s' exceeds the %d byte limit for variable names in netCDF-4 classic model and will be clipped\r",strMaxLen,Wvstr 
			getError("QLA_put_dim",errorStr,2,fileID)
			WvStr=CleanUpName(WvStr,1,strMaxLen-1)
		endif
	endif

	// make a holding wave for the _Netcdf4Coordinates attribute
	Make /O /N=(j) /FREE attWv

	// attach the dimesnions listed to the dataset, get the dimID if required 
	for (i=0;i<j;i+=1)
		dimPathStr = StringFromList(i,dimPathList)
		if (classic)
			dimPathStr = ParseFilePath(0,dimPathStr,"/",1,0) 
		endif
		
		// check that the dimension dataset exists
		if (!existsInFile(ParseFilePath(0,dimPathStr,"/",1,0),fileID))
			sprintf errorStr,"No such dataset named '%s' found in file while attempting to attach it as a dimension of '%s'\r",ParseFilePath(0,dimPathStr,"/",1,0),WvStr
			getError("QLA_put_dim",errorStr,3,fileID)
			continue
		endif
		
		// check that data and dimension are relative to each other
		if (!isRelative(path,dimPathStr))
			sprintf errorStr,"Dimension '%s' is not relative to '%s' in the hierarchy\r",ParseFilePath(0,dimPathStr,"/",1,0),WvStr
			getError("QLA_put_dim",errorStr,3,fileID)
			return -1
		endif
		
		// check that the dimension is a valid netCDF-4 dimension
		HDF5LoadData /Q /Z /O /A="_Netcdf4Dimid" /TYPE=2 /N=tempWv, fileID, dimPathStr
		Wave tempWv ; attWv[i]=tempWv[inf] ; KillWaves/Z tempWv
		
		// handle if the dimension doesn't exist yet
		if (V_flag != 0)
			sprintf errorStr,"No such dimension named '%s' found in file\r",ParseFilePath(0,dimPathStr,"/",1,0)
			getError("QLA_put_dim",errorStr,3,fileID)
			continue
		endif
		
		// handle dimension size miss-match
		if (!isDimValid(dimPathStr,DimSize(Wv,i),fileID))
			sprintf errorStr,"Dimension size miss-match for dimension %d of '%s'\r",i,Wvstr 
			getError("QLA_put_dim",errorStr,3,fileID)
			return -1
		endif
		
		// check that there aren't already dimensions attached
		HDF5DimensionScale /Z dataset={fileID,path}, dimIndex=i, getNumScales
		if (V_value != 0)
			HDF5DimensionScale /Z dataset={fileID,path}, dimIndex=i, getScaleList
			for (l=0;l<ItemsInList(S_Value,"\r");l+=1)
				scalePath=StringFromList(l,S_value,"\r")
				HDF5DimensionScale /Z scale={fileID,scalePath}, dataset={fileID,path} dimIndex=i, detachScale
				sprintf errorStr,"'%s' has been removed as a dimension of '%s' in order to attached '%s' as dimension %d\r",ParseFilePath(0,scalePath,"/",1,0),WvStr,ParseFilePath(0,dimPathStr,"/",1,0),i
				getError("QLA_put_dim",errorStr,2,fileID)
			endfor
		endif 
				
	
		// attach the dimension	
		HDF5DimensionScale /Z scale={fileID,dimPathStr}, dataset={fileID,path}, dimIndex=i, attachScale
	endfor

	// handle errors from HDF5DimesnionScale
	if (V_flag != 0)
		sprintf errorStr,"There was a problem attaching dimensions to the dataset '%s'\r",WvStr
		getError("QLA_put_dim",errorStr,3,fileID)
		return -1
	endif
	
	// write the _Netcdf4Coordinate attribute for multidimensional datasets 
	if (j>1)
		QLA_put_att_wave(path,"_Netcdf4Coordinates",attWv,32,fileID)
	endif
	
	// get an error string to display 
	sprintf errorStr,"%d dimensions attached to '%s'\r",j,WvStr 
	getError("QLA_put_dim",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End	
//___________________________________________________________________________________________________	
// write variable and attach netCDF-4 dimensions to a variable 
Function QLA_def_var(Wave Wv,String path,String dimPathList,int type,int deflate,int shuffle,double fileID [,int storage,Wave chunksizes,int extend])
	
	int i,j=ItemsInList(dimPathList),dims=WaveDims(Wv),l,classic,WvType=WaveType(Wv),chunk,error
	String dimPathStr, WvStr = ParseFilePath(0,path,"/",1,0),errorStr
		 
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// set the append option correctly 
	extend = (ParamIsDefault(extend)) ? -1 : extend

	// is the name of the variable the same as one of the coordinates?
	error = sameNameAsDim(path,fileID)
	if (error == 2) // same name as a dimension
		path = ParseFilePath(1,path,"/",1,0)+"_nc4_non_coord_"+WvStr
	elseif (error == 1) // same name as a dimension variable
		sprintf errorStr,"'%s' already exists as a dimension variable, cannot create variable of the same name\r",WvStr
		getError("QLA_def_var",errorStr,3,fileID)
		return -1
	endif
	
	// check the name for CF compliance
	if (!isCFName(WvStr))
		sprintf errorStr,"The variable name '%s' is allowed in netCDF-4 but is not a CF compliant name\r",WvStr
		getError("QLA_def_var",errorStr,2,fileID)
	endif

	// check the number of dimensions is correct
	if (dims != j)
		sprintf errorStr,"Number of dimensions in the dataset '%s' (%d) does not match the number of dimensions to attach (%d)\r",WvStr,dims,j
		getError("QLA_def_var",errorStr,3,fileID)
		return -1 
	endif	
	
	// create local copy of data in case it must be redimensioned
	Duplicate /O /FREE Wv outWv	
	
	// change the wavetype if requested
	if ((type > 1) && (Wvtype != type))
		Redimension /Y=(type) outWv
		WvType = type
	endif 
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		classic=1 ;	path = WvStr // make sure the path refers to root
		
		// make sure the variable name isn't too long
		if (strlen(WvStr) > strMaxLen)
			sprintf errorStr,"The name '%s' exceeds the %d byte limit for variable names in netCDF-4 classic model and will be clipped\r",strMaxLen,Wvstr 
			getError("QLA_def_var",errorStr,2,fileID)
			WvStr=CleanUpName(WvStr,1,strMaxLen-1)
		endif
		
		// warn if we are overwriting something (unintentionally) - possible if code was written for groups and then classic model was passed 
		if (sameName(path,fileID) && (extend == -1))
			sprintf errorStr,"The variable '"+WvStr+"' already exists in the output file so the data will be overwritten\r"
			getError("QLA_def_var",errorStr,2,fileID)
		endif
		
		// warn for illegal data types, commute if necessary
		if (WvType == 0)
			sprintf errorStr,"'%s' is of a string data type and is illegal in netCDF-4 classic model\r",WvStr
			getError("QLA_def_var",errorStr,3,fileID)
			return -1
		elseif (WvType > 32)
			type = isClassicDataType(WvType)
			sprintf errorStr,"'%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",WvStr,waveType2String(type)
			getError("QLA_def_var",errorStr,1,fileID)
			Redimension /Y=(type) outWv
		endif
	endif
	
	// make a holding wave for the _Netcdf4Coordinates attribute
	Make /O /N=(j) /Y=32 /FREE attWv
	
	// make a holding wave for the dimension sizes
	Make /O /N=4 /Y=128 /FREE dimsWv

	// check the validity of the dimensions 
	for (i=0;i<j;i+=1)
		dimPathStr = StringFromList(i,dimPathList)
		if (classic)
			dimPathStr = ParseFilePath(0,dimPathStr,"/",1,0) 
		endif
		
		// check that the dimension dataset exists
		if (!existsInFile(ParseFilePath(0,dimPathStr,"/",1,0),fileID))
			sprintf errorStr,"No such dataset named '%s' found in file while attempting to attach it as a dimension of '%s'\r",ParseFilePath(0,dimPathStr,"/",1,0),WvStr
			getError("QLA_def_var",errorStr,3,fileID)
			continue
		endif
		
		// check that data and dimension are relative to each other
		if (!isRelative(path,dimPathStr))
			sprintf errorStr,"Dimension '%s' is not relative to '%s' in the file hierarchy\r",ParseFilePath(0,dimPathStr,"/",1,0),WvStr
			getError("QLA_def_var",errorStr,3,fileID)
			return -1
		endif
		
		// check that the dimension is a valid netCDF-4 dimension
		HDF5LoadData /Q /Z /O /A="_Netcdf4Dimid" /TYPE=2 /N=tempWv, fileID, dimPathStr
		Wave tempWv ; attWv[i]=tempWv[inf] ; KillWaves/Z tempWv
				
		// handle if the dimension doesn't exist yet
		if (V_flag != 0)
			sprintf errorStr,"No such dimension named '%s' found in file\r",ParseFilePath(0,dimPathStr,"/",1,0)
			getError("QLA_def_var",errorStr,3,fileID)
			continue
		endif
		
		// get the dimension sizes/unlimited dimensions
		dimsWv[i]=getMaxD(dimPathStr,0,fileID)
		
		// handle dimension size miss-match if dataset isn't being extended or has no unlimited dimension
		if (!isDimValid(dimPathStr,DimSize(Wv,i),fileID) && (dimsWv[i] > -1))
			sprintf errorStr,"Dimension size miss-match for dimension %d of '%s'\r",i,Wvstr 
			getError("QLA_def_var",errorStr,2,fileID)
		endif
	endfor
	
	// check that if there is an unlimited dimension it is attached to dim0 as required in classic model
	if (classic && (WaveMin(dimsWv) == -1) && (dimsWv[0] != -1))
		sprintf errorStr,"The unlimited dimension of '%s' is not at index 0 as required by netCDF-4 classic model\r" WvStr
		getError("QLA_def_var",errorStr,2,fileID)
	endif
	
	// does the data need chunking
	chunk = (extend > -1) || (WaveMin(dimsWv) == -1) || (deflate > 0) || (shuffle == 1) || (storage == 1) || (ParamIsDefault(storage) && (NumberByKey("SIZEINBYTES",WaveInfo(outWv,0)) > contiguousMaxBytes)) ? 1 : 0
	
	// get some chunksizes
	Wave chunkWv = getChunkSizes(outWv,dimsWv)
	
	if (!ParamIsDefault(chunksizes))
		l = numpnts(chunksizes)
		if (l == dims)
			chunkWv[0,l-1] = chunksizes[p]
		else
			sprintf errorStr,"The number of chunks specified (%d) does not match the number of dimensions (%d) of '%s'\r",l,dims,WvStr
			getError("QLA_def_var",errorStr,2,fileID)
		endif
	endif
	
	// the compression parameters if the dataset exists
	if (existsInFile(path,fileID))
		
		// get the compression type of the data stored on disk
		[chunk,deflate,shuffle,chunksizes] = compOnDisk(path,fileID)
			
		// data is contiguous 
		if (!chunk) 
			sprintf errorStr,"The existing dataset '%s' is stored contiguously and cannot be extended, use chunked layout for extensible datasets\r",WvStr
			getError("QLA_def_var",errorStr,3,fileID)
			return 0
		endif
		
		// check the data type
		type = typeOnDisk(path,fileID)
		if ((WvType != type) && (type > 1) && (WvType > 1)) // both numeric but a different type
			Redimension /Y=(type) outWv
		elseif ((WvType != type) && ((type < 2) || (WvType < 2))) // one is string
			sprintf errorStr,"Cannot concatenate %s and %s types together\r",waveType2String(type),waveType2String(WvType)
			getError("QLA_def_var",errorStr,3,fileID)
			return 0
		endif
	endif  
	
	// write the data to file
	HDF5SaveData /O /Z /IGOR=0 /OPTS=1 /APND=(extend) /LAYO={2^chunk,chunkWv[0],chunkWv[1],chunkWv[2],chunkWv[3]} /GZIP={deflate,shuffle} /MAXD={dimsWv[0],dimsWv[1],dimsWv[2],dimsWv[3]} outWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0)
		sprintf errorStr,"There was a problem creating the dataset '%s'\r",WvStr
		getError("QLA_def_var",errorStr,3,fileID)
		return -1
	endif

	// attach the dimensions
	for (i=0;i<j;i+=1)
		dimPathStr = StringFromList(i,dimPathList)
		if (classic)
			dimPathStr = ParseFilePath(0,dimPathStr,"/",1,0) 
		endif	
		
		HDF5DimensionScale /Z scale={fileID,dimPathStr}, dataset={fileID,path}, dimIndex=i, attachScale
	endfor
	
	// handle errors from HDF5DimesnionScale
	if (V_flag != 0)
		sprintf errorStr,"There was a problem attaching dimensions to the dataset '%s'\r",WvStr
		getError("QLA_def_var",errorStr,3,fileID)
		return -1
	endif
 
 	// write the _Netcdf4Coordinate attribute for multidimensional datasets 
	if (j > 1)
		QLA_put_att_wave(path,"_Netcdf4Coordinates",attWv,32,fileID)
	endif

	// get an error string to display 
	sprintf errorStr,"'%s' variable created\r",WvStr 
	getError("QLA_def_var",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// function to 'update' the data in a variable on disk without having to redefine all the dimensions and attributes
Function QLA_update_var(Wave Wv,String Path,double fileID)
	
	DFREF dfrSave = GetDataFolderDFR()
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif

	SetDataFolder NewFreeDataFolder()	// Create new free data folder
	
	int i,type ; String attStr,dimPathLst,WvStr = ParseFilePath(0,path,"/",1,0)
	String errorStr
	
	// check if this netCDF is classic model
	if (isClassic(fileID))
			// make sure the variable name isn't too long
		if (strlen(WvStr) > strMaxLen)
			WvStr=CleanUpName(WvStr,1,strMaxLen-1)
		endif
		
		path = WvStr // make sure the path refers to root
	endif
	
	// does the dataset exist
	if (!existsInFile(path,fileID))
		SetDataFolder dfrSave
		sprintf errorStr,"No such variable named '%s' found while attempting to update the dataset\r",WvStr
		getError("QLA_update_var",errorStr,3,fileID)		
		return 0
	endif
	
	// get the datatype of the variable stored on disk
	type = typeOnDisk(path,fileID)
	
	// get the compression type of the data stored on disk
	[int chunk,int deflate,int shuffle,Wave chunksizes] = compOnDisk(path,fileID)
	
	// get a list of dimensions of the variable stored on disk
	dimPathLst = getHDF5Dimensions(fileID,path) 

	// get a list of attributes of teh variable stored on disk
	HDF5ListAttributes /TYPE=2 /Z fileID, path
	
	// remove HDF5 attribute we can't write from list 
	S_HDF5ListAttributes = RemoveFromList(HDF5markup,S_HDF5ListAttributes)
	
	// load the attributes to a free datafolder
	for (i=0;i<ItemsInList(S_HDF5ListAttributes);i+=1)
		attStr = StringFromList(i,S_HDF5ListAttributes)
		HDF5LoadData /O /A=attStr /TYPE=2 /OPTS=1 /IGOR=0 /N=$"temp_"+attStr /Q /Z fileID, path
	endfor
	
	// write the new data to disk 
	QLA_def_var(Wv,path,dimPathLst,type,deflate,shuffle,fileID,storage=chunk,chunksizes=chunksizes)
	
	// attach the saved attribues to the new data
	for (i=0;i<ItemsInList(S_HDF5ListAttributes);i+=1)
		attStr = StringFromList(i,S_HDF5ListAttributes)
		// string
		if (WaveType($"temp_"+attStr) == 0)
			Wave/Z/T AttStrWv = $"temp_"+attStr
			QLA_put_att_text(path,attStr,AttStrWv[inf],fileID)
		// numeric
		elseif (WaveType($"temp_"+attStr) != 0)
			Wave/Z AttVarWv = $"temp_"+attStr
			QLA_put_att_wave(path,attStr,AttVarWv,WaveType(AttVarWv),fileID)
		endif
	endfor

	SetDataFolder dfrSave
	
	// get an error string to display 
	sprintf errorStr,"'%s' variable updated\r",WvStr 
	getError("QLA_update_var",errorStr,0,fileID)
	
	return 1
End
//___________________________________________________________________________________________________	
// extend an existing dataset or co-ordinate
Function QLA_extend(Wave Wv,String Path,int extend,double fileID)
	
	String WvStr = ParseFilePath(0,path,"/",1,0),errorStr ; int type, WvType=WaveType(Wv)

	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check if this netCDF is classic model
	if (isClassic(fileID))
			// make sure the variable name isn't too long
		if (strlen(WvStr) > strMaxLen)
			WvStr=CleanUpName(WvStr,1,strMaxLen-1)
		endif
		
		path = WvStr // make sure the path refers to root
	endif

	// the compression parameters if the dataset exists
	if (existsInFile(path,fileID))
		
		// create local copy of data in case it must be redimensioned
		Duplicate /O /FREE Wv outWv
		
		// get the compression type of the data stored on disk
		[int chunk,int deflate,int shuffle,Wave chunkWv] = compOnDisk(path,fileID)
			
		// data is contiguous 
		if (!chunk) 
			sprintf errorStr,"The existing dataset '%s' is stored contiguously and cannot be extended, use chunked layout for extensible datasets\r",WvStr
			getError("QLA_extend",errorStr,3,fileID)
			return 0
		endif
	
		// check the data type
		type = typeOnDisk(path,fileID)
		if ((WvType != type) && (type > 1) && (WvType > 1)) // both numeric but a different type
			Redimension /Y=(type) outWv
		elseif ((WvType != type) && ((type < 2) || (WvType < 2))) // one is string
			sprintf errorStr,"Cannot concatenate char and numeric types together\r",waveType2String(type),waveType2String(WvType)
			getError("QLA_extend",errorStr,3,fileID)
			return 0
		endif
		
		// check that the dimension is extensible
		if (getMaxD(path,extend,fileID) != -1)
			sprintf errorStr,"Dimension %d of '%s' is of finite size and cannot be extended\r",extend,WvStr
			getError("QLA_extend",errorStr,3,fileID)
			return 0
		endif
	else
		sprintf errorStr,"No such variable named '%s' found while attempting to extend the dataset\r",WvStr
		getError("QLA_extend",errorStr,3,fileID)
		return 0
	endif
	
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /APND=(extend) /LAYO={2^chunk,chunkWv[0],chunkWv[1],chunkWv[2],chunkWv[3]} /GZIP={deflate,shuffle} outWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0)
		sprintf errorStr,"There was a problem extending the dataset '%s'\r",WvStr
		getError("QLA_extend",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' variable extended\r",WvStr 
	getError("QLA_extend",errorStr,0,fileID)
	
	return 1
End
//___________________________________________________________________________________________________	
// write netCDF-4 string attribute to a variable
Function QLA_put_att_text(String path,String attName,String attValue,double fileID)
	
	String WvStr = ParseFilePath(0,path,"/",1,0),errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check if name is the same as a dimension
	if (sameNameAsDim(path,fileID) == 2)
		path=ParseFilePath(1,path,"/",1,0)+"_nc4_non_coord_"+WvStr
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		// make sure the path refers to root
		path = WvStr 
		
		// make sure the variable name isn't too long
		WvStr=CleanUpName(WvStr,1,strMaxLen-1)
	endif

	// check to see if the dataset exists
	if (!existsInFile(WvStr,fileID))
		sprintf errorStr,"No such dataset named '%s' found in file\r",WvStr
		getError("QLA_put_att_text",errorStr,3,fileID)
		return -1
	endif
	
	// check the size of the attribute
	if (strlen(attValue) > attMaxBytes)
		sprintf errorStr,"The '%s' attribute for the variable '%s' exceeds the maximum number of bytes allowed (%d) and has been clipped\r",attName,WvStr,attMaxBytes
		getError("QLA_put_att_text",errorStr,2,fileID)
		attValue = attValue[0,attMaxBytes-1]
	elseif (strlen(attValue) == 0)
		return 0
	endif

	// write the attribute
	Make /O /T /FREE /N=1 StrAttWv = attValue
	HDF5SaveData /Z /IGOR=0 /O /A=attName StrAttWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' attribute for the variable '%s'\r",attName,WvStr
		getError("QLA_put_att_text",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' attribute set for for the variable '%s'\r",attName,WvStr
	getError("QLA_put_att_text",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1 
End
//___________________________________________________________________________________________________	
// write netCDF-4 variable list attribute to a variable from a string list
Function QLA_put_att_list(String path,String attName,String attValue,String attSep,int type,double fileID)
	
	int i,j = ItemsInList(attValue) ; String WvStr = ParseFilePath(0,path,"/",1,0),errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
		
	// make sure a string attribute isn't being requested
	if (type == 0)
		sprintf errorStr,"Attempted to attach '%s' as a text attribute to '%s' using QLA_put_att_list; use e.g. QLA_put_att_text or set type to a non-zero value\r",attName,wvStr
		getError("QLA_put_att_list",errorStr,3,fileID)
		return 0
	endif
	
	// check if name is the same as a dimension
	if (sameNameAsDim(path,fileID) == 2)
		path=ParseFilePath(1,path,"/",1,0)+"_nc4_non_coord_"+WvStr
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		// make sure the path refers to root
		path = WvStr 
		
		// check for illegal data types and commute as needed
		if (type > 32)
			type = isClassicDataType(type)
			sprintf errorStr,"'%s' attribute attached to '%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",attName,wvStr,waveType2String(type)
			getError("QLA_put_att_list",errorStr,1,fileID)
		endif
	endif
	
	// check to see if the data set exists
	if (!existsInFile(WvStr,fileID))
		sprintf errorStr,"No such dataset named '%s' found in file\r",WvStr
		getError("QLA_put_att_list",errorStr,3,fileID)
		return -1
	endif
	
	// set the attribute format to the same as the variable if it is respected
	type = respectedAttribute(path,attName,type,fileID)
	
	// make a holding wave for the list attribute 
	Make /O /N=(j) /Y=(type) /FREE attWv = str2num(StringFromList(p,attValue,attSep))

	// check the size of the attribute
	if (NumberByKey("SIZEINBYTES",WaveInfo(attWv,0)) > attMaxBytes)
		sprintf errorStr,"The '%s' attribute for the variable '%s' exceeds the maximum number of bytes allowed (%d)\r",attName,WvStr,attMaxBytes
		getError("QLA_put_att_list",errorStr,3,fileID)
		return -1
	endif

	// write the attribute
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /O /A={attName,1} attWv, fileID, path
	
	 // handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' attribute for the variable '%s'\r",attName,WvStr
		getError("QLA_put_att_list",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' attribute set for for the variable '%s'\r",attName,WvStr
	getError("QLA_put_att_list",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// write netCDF-4 variable list attribute to a variable from a wave
Function QLA_put_att_wave(String path,String attName,Wave attValue,int type,double fileID)
	
	int j = numpnts(attValue),WvType=WaveType(attValue) ; String WvStr = ParseFilePath(0,path,"/",1,0),errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check the type of the wave was set 
	type = (type == 0) ? WvType : type
	
	// check if name is the same as a dimension
	if (sameNameAsDim(path,fileID) == 2)
		path=ParseFilePath(1,path,"/",1,0)+"_nc4_non_coord_"+WvStr
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		// make sure the path refers to root
		path = WvStr 
		
		// check for illegal data types and commute as needed
		if (type > 32)
			type = isClassicDataType(type)
			sprintf errorStr,"'%s' attribute attached to '%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",attName,wvStr,waveType2String(type)
			getError("QLA_put_att_wave",errorStr,1,fileID)
		endif
	endif
	
	// check to see if the data set exists
	if (!existsInFile(WvStr,fileID))
		sprintf errorStr,"No such dataset named '%s' found in file\r",WvStr
		getError("QLA_put_att_wave",errorStr,3,fileID)
		return -1
	endif
		
	// set the attribute format to the same as the variable if it is respected
	type = respectedAttribute(path,attName,type,fileID)
		
	// make the attribute
	Make /O /N=(j) /Y=(type) /FREE attWv = attValue
	
	// check the size of the attribute
	if (NumberByKey("SIZEINBYTES",WaveInfo(attWv,0)) > attMaxBytes)
		sprintf errorStr,"The '%s' attribute for the variable '%s' exceeds the maximum number of bytes allowed (%d)\r",attName,WvStr,attMaxBytes
		getError("QLA_put_att_wave",errorStr,3,fileID)
		return -1
	endif
	
	// write the attribute
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /O /A={attName,1} attWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' attribute for the variable '%s'\r",attName,WvStr
		getError("QLA_put_att_wave",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' attribute set for for the variable '%s'\r",attName,WvStr
	getError("QLA_put_att_wave",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// write netCDF-4 variable attribute to a variable
Function QLA_put_att(String path,String attName,Variable attValue,int type,double fileID)
	
	String WvStr = ParseFilePath(0,path,"/",1,0),errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// make sure a string attribute isn't being requested
	if (type == 0)
		sprintf errorStr,"Attempted to attach '%s' as a text attribute to '%s' using QLA_put_att; use e.g. QLA_put_att_text or set type to a non-zero value\r",attName,wvStr
		getError("QLA_put_att",errorStr,3,fileID)
		return 0
	endif
	
	// check if name is the same as a dimension
	if (sameNameAsDim(path,fileID) == 2)
		path = ParseFilePath(1,path,"/",1,0)+"_nc4_non_coord_"+WvStr
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		// make sure the path refers to root
		path = WvStr 
		
		// make sure the variable name isn't too long
		WvStr = CleanUpName(WvStr,1,strMaxLen-1)
		
		// check for illegal data types and commute as needed
		if (type > 32)
			type = isClassicDataType(type)
			sprintf errorStr,"'%s' attribute attached to '%s' is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",attName,wvStr,waveType2String(type)
			getError("QLA_put_att",errorStr,1,fileID)
		endif
	endif

	// check to see if the data set exists
	if (!existsInFile(WvStr,fileID))
		sprintf errorStr,"No such dataset named '%s' found in file\r",WvStr
		getError("QLA_put_att",errorStr,3,fileID)
		return -1
	endif
	
	// set the attribute format to the same as the variable if it is respected
	type = respectedAttribute(path,attName,type,fileID)
	
	// write the attribute
	Make /O /FREE /N=1 /Y=(type) attWv = attValue
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /O /A={attName,1} attWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' attribute for the variable '%s'\r",attName,WvStr
		getError("QLA_put_att",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' attribute set for for the variable '%s'\r",attName,WvStr
	getError("QLA_put_att",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// write netCDF group/global string attribute
Function QLA_put_glob_att_text(String path,String attName,String attValue,double fileID)
	
	String errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		
		// groups likely contain the same global attributes, preserve this information 
		// flatten the hierarchy into a single string and prepend to the attribute name
		attName = CleanupName(RemoveEnding(path,"/")+"/",0)[2,inf]+attName
		
		// make sure the path refers to root
		path = "/" 
	endif
	
	// check the size of the attribute
	if (strlen(attValue) > attMaxBytes)
		sprintf errorStr,"The '%s' global attribute exceeds the maximum number of bytes allowed (%d) and has been clipped\r",attName,attMaxBytes
		getError("QLA_put_glob_att_text",errorStr,2,fileID)
		attValue = attValue[0,attMaxBytes-1]
	elseif (strlen(attValue) == 0)
		return 0
	endif
	
	// write the attribute
	Make /FREE /T /N=1 AttWv = attValue
	HDF5SaveData /Z /IGOR=0 /O /A=attName attWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' global attribute \r",attName
		getError("QLA_put_glob_att_text",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' global attribute set\r",attName
	getError("QLA_put_glob_att_text",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// write netCDF-4 group/global variable attribute 
Function QLA_put_glob_att(String path,String attName,Variable attValue,int type,double fileID)

	String errorStr

	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// make sure a string attribute isn't being requested
	if (type == 0)
		sprintf errorStr,"Attempted to create '%s' as a global text attribute using QLA_put_glob_att; use e.g. QLA_put_glob_att_text or set type to a non-zero value\r",attName
		getError("QLA_put_glob_att",errorStr,3,fileID)
		return 0
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
	
		// groups likely contain the same global attributes, preserve this information 
		// flatten the hierarchy into a single string and prepend to the attribute name
		attName = CleanupName(RemoveEnding(path,"/")+"/",0)[2,inf]+attName
		
		// make sure the path refers to root
		path = "/" 
		
		// check for illegal data types and commute as needed
		if (type > 32)
			type = isClassicDataType(type)
			sprintf errorStr,"'%s' global attribute is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",attName,waveType2String(type)
			getError("QLA_put_glob_att",errorStr,1,fileID)
		endif
	endif

	// write the attribute
	Make /O /FREE /N=1 /Y=(type) attWv = attValue
	HDF5SaveData /IGOR=0 /O /OPTS=1 /A={attName,1} attWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' global attribute \r",attName
		getError("QLA_put_glob_att",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' global attribute set\r",attName
	getError("QLA_put_glob_att",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// write netCDF-4 wave global attribute
Function QLA_put_glob_att_wave(String path,String attName,Wave attValue,int type,double fileID)
	
	double j = numpnts(attValue) ; String errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		// groups likely contain the same global attributes, preserve this information 
		// flatten the hierarchy into a single string and prepend to the attribute name
		attName = CleanupName(RemoveEnding(path,"/")+"/",0)[2,inf]+attName
		
		// make sure the path refers to root
		path = "/" 
		
		// check for illegal data types and commute as needed
		if (type > 32)
			type = isClassicDataType(type)
			sprintf errorStr,"'%s' global attribute is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",attName,waveType2String(type)
			getError("QLA_put_glob_att_wave",errorStr,1,fileID)
		endif
	endif
		
	// make the attribute
	Make /O /N=(j) /Y=(type) /FREE attWv = attValue
	
	// check the size of the attribute
	if (NumberByKey("SIZEINBYTES",WaveInfo(attWv,0)) > attMaxBytes)
		sprintf errorStr,"The '%s' global attribute ' exceeds the maximum number of bytes allowed (%d)\r",attName,attMaxBytes
		getError("QLA_put_glob_att_wave",errorStr,3,fileID)
		return -1
	endif
	
	// write the attribute
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /O /A={attName,1} attWv, fileID, path
	
	// handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' global attribute for the variable '%s'\r",attName
		getError("QLA_put_glob_att_wave",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' global attribute set\r",attName
	getError("QLA_put_glob_att_wave",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// write netCDF-4 variable list global attribute 
Function QLA_put_glob_att_list(String path,String attName,String attValue,String attSep,int type,double fileID)
	
	int j = ItemsInList(attValue) ; String errorStr
	
	// check the status of the file
	if (!fileStatus(fileID))
		return 0
	endif
	
	// check if this is a classic model netCDF
	if (isClassic(fileID))
		// groups likely contain the same global attributes, preserve this information 
		// flatten the hierarchy into a single string and prepend to the attribute name
		attName = CleanupName(RemoveEnding(path,"/")+"/",0)[2,inf]+attName
		
		// make sure the path refers to root
		path = "/" 
		
		// check for illegal data types and commute as needed
		if (type > 32)
			type = isClassicDataType(type)
			sprintf errorStr,"'%s' global attribute is of an illegal data type in netCDF-4 classic model and has been commuted to '%s'\r",attName,waveType2String(type)
			getError("QLA_put_glob_att_list",errorStr,1,fileID)
		endif
	endif

	// make a holding wave for the list attribute 
	Make /O /N=(j) /Y=(type) /FREE attWv = str2num(StringFromList(p,attValue,attSep))

	// check the size of the attribute
	if (NumberByKey("SIZEINBYTES",WaveInfo(attWv,0)) > attMaxBytes)
		sprintf errorStr,"The '%s' global attribute exceeds the maximum number of bytes allowed (%d)\r",attName,attMaxBytes
		getError("QLA_put_glob_att_list",errorStr,3,fileID)
		return -1
	endif

	// write the attribute
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /O /A={attName,1} attWv, fileID, path
	
	 // handle errors from HDF5SaveData
	if (V_flag != 0) 
		sprintf errorStr,"There was a problem writing the '%s' global attribute\r",attName
		getError("QLA_put_glob_att_list",errorStr,3,fileID)
		return -1
	endif
	
	// get an error string to display 
	sprintf errorStr,"'%s' global attribute set\r",attName
	getError("QLA_put_glob_att_list",errorStr,0,fileID)
	
	HDF5FlushFile fileID
	return 1
End
//___________________________________________________________________________________________________	
// close the netCDF if it is open
Function QLA_close(double fileID)

	String errorStr

	// check the status of the file
	if (!fileStatus(fileID))
		sprintf errorStr,"There was a problem closing the file (id: %d)\r",fileID
		getError("QLA_close",errorStr,3,fileID)
		return 0
	else
		// flush the file
		HDF5FlushFile /Z fileID
		// close the file
		HDF5CloseFile / Z fileID
	endif
	
	// get an error string to display 
	sprintf errorStr,"netCDF-4 closed (id: %d)\r",fileID
	getError("QLA_close",errorStr,0,fileID)
	
	return 1
End
//___________________________________________________________________________________________________	
// get some _NCProperties that might be required in netCDF APIs, hdf5 version correct from igor  
// _NCProperties is not strictly required as the _IsNetcdf4 attribute is computed anyway by the API
// by convention, but not enforced, in version=2 the next key pair would be the netCDF library version
// e.g. netcdf=4.7.4 - since we don't use any netCDF library it's ommited, would be netcdf=0.0.0 maybe
// format verion=2,qla=N.N.N,hdf5=N.NN.NN,igor=N.N.N.N
Static Function/S NCProperties()

	String rtnStr="version=2",cmdstr
	String hdf5libVer = num2str(HDF5LibraryInfo(0))
	String qlalibver = num2str(QLA_version)
	String igorlibVer = ",igor="+StringByKey("IGORFILEVERSION",IgorInfo(3))
	
	int major,minor,release
	 
	sscanf qlalibver,"%1d%1d%1d",major,minor,release
	sprintf cmdstr,",qla=%d.%d.%d",major,minor,release
	rtnStr+=cmdstr
	sscanf hdf5libVer,"%1d%2d%2d",major,minor,release
	sprintf cmdstr,",hdf5=%d.%d.%d",major,minor,release
	rtnStr+=cmdstr+igorlibVer
	
	return rtnStr
End
//___________________________________________________________________________________________________	
// check the status of a file
Static Function fileStatus(double fileID)
	double groupID
	HDF5OpenGroup /Z fileID,".",groupID
	if (V_flag == 0)
		return 1
	endif
	return 0
End
//___________________________________________________________________________________________________	
// get the next free dimension ID number - they are arbitrary but must be unique and monotonic
Static Function getNextNCdimID(double fileID)
	
	int i,j,k =-1 ; String dimPathStr
	
	// get a list of all the variables in the file
	HDF5ListGroup /R fileID, "/"
	
	j = ItemsInList(S_HDF5ListGroup)
	
	// count how many variables have the _Netcdf4Dimid attribute
	for (i=0;i<j;i+=1)
		HDF5ListAttributes /Type=2 fileID, StringFromList(i,S_HDF5ListGroup)
		if (GrepString(S_HDF5ListAttributes,"_Netcdf4Dimid")) 
			k+=1 
		endif
	endfor
	
	return k+1
End
//___________________________________________________________________________________________________	
// test to see if a dataset is named the same as an existing one in the same group
Static Function sameName(String path,double fileID)

	int i,j ; String Wv2Str

	// get a list of all variables in the same group 
	HDF5ListGroup fileID, "/"+ParseFilePath(1,path,"/",1,0)
	
	j = ItemsInList(S_HDF5ListGroup)
	
	// compare names of all variables in the group as the one we want to write
	for (i=0;i<j;i+=1)
		Wv2Str = StringFromList(i,S_HDF5ListGroup)
		if (CmpStr(path,Wv2Str)==0)
			return 1
		endif
	endfor
	
	return 0
End
//___________________________________________________________________________________________________	
// test to see if a dataset is named the same as an existing variable for the rare occurrence that it is
// avoid this firstly, secondly prepend _nc4_non_coord_ to the variable name - it is hidden by the API
Static Function sameNameAsDim(String path,double fileID)

	// get a list of all the variables in the file
	HDF5ListGroup /R fileID, "/"
	
	int i,j = ItemsInList(S_HDF5ListGroup) ; String Wv1Str,Wv2Str,dimPathStr
	
	Wv1Str = ParseFilePath(0,path,"/",1,0)
	
	// compare names of all dimensions in the file to the one we want to write
	for (i=0;i<j;i+=1)
		dimPathStr = StringFromList(i,S_HDF5ListGroup)
		Wv2Str = ParseFilePath(0,dimPathStr,"/",1,0)
		HDF5LoadData /Q /Z /O /A="_Netcdf4Dimid" /TYPE=2 /N=tempWv ,fileID ,dimPathStr
		KillWaves /Z tempWv
		if (V_flag==0 && CmpStr(Wv1Str,Wv2Str)==0)
				
			// test to see if it is a dimension but not a variable
			HDF5LoadData /Q /Z /O /A="NAME" /TYPE=2 /N=tempWv ,fileID ,dimPathStr
			Wave/T tempWV
			if (StringMatch(tempWv[inf],"This is a netCDF dimension but not a netCDF variable.*"))
				KillWaves /Z tempWv
				return 2
			endif
			KillWaves /Z tempWv
			return 1 // the variable has the same name as a dimension
		endif
	endfor
	
	return 0
End
//___________________________________________________________________________________________________	
// test to see if file is netCDF4 'classic model' indicated by the presence of _nc3_strict attribute
// the only valid value of _nc3_strict is 1, no need to check there is no _nc3_strict=0
Static Function isClassic(double fileID)

	// look for the _nc3_strict attribute in the root of the file
	HDF5ListAttributes /Type=1 fileID, "/"
	if (GrepString(S_HDF5ListAttributes,"_nc3_strict"))
		return 1
	endif
	
	return 0
End
//___________________________________________________________________________________________________	
// check data type and commute illegal numeric types to ones which can best represent the possible maximum value
Static Function isClassicDataType(double type)

	int classicType
	
	// switch on datatypes not allowed in netCDF-3/netCDF-4 classic model
	switch (type)
		case 0:		// string
			classictype = 1 	// char
			break 
		case 72:	// ubyte
			classicType = 16	// short
			break
		case 80:	// ushort
			classicType = 32	// long	
			break	
		case 96:	// uint	
			classicType = 4		// double
			break
		case 128:	// int64
			classicType = 4		// double
			break
		case 192:	// uint64
			classicType = 4		// double
			break
		default:
			classicType = type
	endswitch
	
	return classicType
End
//___________________________________________________________________________________________________	
// compare the size of the co-ordinate to the size of the dimension we are trying to attach it to 
Static Function isDimValid(String dimPathStr,double varDimLen,double fileID)
	
	int dummy
	double coordDimLen
	
	STRUCT HDF5DataInfo di	
	InitHDF5DataInfo(di)	
	dummy = HDF5DatasetInfo(fileID, dimPathStr, 0, di)
	coordDimLen = di.dims[0]
	
	if (coordDimLen > varDimLen)
		return 0 // not valid 
	endif
	
	return 1
End
//___________________________________________________________________________________________________	
// get wavetype as a string
Static Function/S waveType2String(int type)
	
	String typeStr = ""

	switch(type)
		case 0: 
			typeStr="string"
			break
		case 1: 
			typeStr="char"	// not a real type
			break
		case 2:
			typeStr="float"
			break
		case 4:
			typeStr="double"
			break
		case 8:
			typeStr="byte"
			break
		case 16:
			typeStr="short"
			break
		case 32:
			typeStr="long" // int
			break
		case 72:  
			typeStr="ubyte"
			break
		case 80:  
			typeStr="ushort"
			break
		case 96:  
			typeStr="uint" // ulong
			break
		case 128:  
			typeStr="int64" // llong
			break
		case 192:  
			typeStr="uint64" // ullONG
			break
	endswitch
	
	return typeStr
End
//___________________________________________________________________________________________________	
// check if the attribute being set is one which the datatype must match the variable
Static Function respectedAttribute(String path,String attName,int type,double fileID)

	int dummy ;	String atts = respectedAttributes
	
	if (GrepString(atts,attName))
	
		STRUCT HDF5DataInfo di	
		InitHDF5DataInfo(di)	
		dummy = HDF5DatasetInfo(fileID, path, 0, di)
		
		// set the attribute format to the same as the variable
		type = h5DataType2WaveType(di.datatype_str)
	endif	
		
	return type 
End
//___________________________________________________________________________________________________	
// get wavetype from HDF5 data type string
Static Function h5DataType2WaveType(String h5DataType)

	int type
	
	strswitch (h5DataType)
		case "string":
			type = 0	// string
			break
		case "32-bit LE float":
			type = 2	// float
			break
		case "64-bit LE float":
			type = 4	// double
			break
		case "8-bit signed LE integer":
			type = 8	// byte
			break
		case "16-bit signed LE integer":
			type = 16	// short
			break
		case "32-bit signed LE integer":
			type = 32	// long / int
			break
		case "8-bit unsigned LE integer":
			type = 72	// ubyte
			break
		case "16-bit unsigned LE integer":
			type = 80	// ushort
			break
		case "32-bit unsigned LE integer":
			type = 96	// uint
			break
		case "64-bit signed LE integer":
			type = 128	// int64 
			break
		case "64-bit unsigned LE integer":
			type = 192	// uint64 
			break	
	endswitch
	
	return type
End 
//___________________________________________________________________________________________________	
// test paths of co-ordinate and variable for hierarchy
Static Function isRelative(String path,String DimPath)

	int i,j,k ; String str1,str2
	
	// get the paths only
	path = ParseFilePath(1,path,"/",1,0)
	DimPath = ParseFilePath(1,DimPath,"/",1,0)
	
	// get hierarchy depth
	j=ItemsInlist(path,"/")
	k=ItemsInlist(DimPath,"/")
	
	// dimension can't be deeper in the hierarchy
	if (j<k)
		return 0 // paths don't match 
	endif
	
	// test paths to see if they match
	for (i=0;i<k;i+=1)
	
		// walk the tree
		str1 = ParseFilePath(0,path,"/",i,0)
		str2 = ParseFilePath(0,DimPath,"/",i,0)

		if (CmpStr(str1,str2,2)==1)
			return 0 // paths don't match
		endif
	endfor
	
	return 1 // paths are relative
End
//___________________________________________________________________________________________________	
// test whether a variable exists in a file
Static Function existsInFile(String path,double FileID)

	int i,J ; String Wv1Str = ParseFilePath(0,path,"/",1,0),Wv2Str
	
	HDF5ListGroup /R fileID, "/" 
	
	j = ItemsInList(S_HDF5ListGroup)
	
	// compare names of all variables in the group as the one we want to write
	for (i=0;i<j;i+=1)
		Wv2Str = ParseFilePath(0,StringFromList(i,S_HDF5ListGroup),"/",1,0)
		if (CmpStr(Wv1Str,Wv2Str)==0)
			return 1
		endif
	endfor
	
	return 0
End
//___________________________________________________________________________________________________	
// test a co-ordinate variable for invalid values and monotonicity
// Wv should be 1D, char should be converted to text before being passed
Static Function [int valid,String error] isValidCoord(Wave Wv)

	uint64 i,j = numpnts(Wv),type=WaveType(Wv)  
	Variable k = Wv[0] 
	
	// check for duplicate values in datasets
	if (type == 0)
		FindDuplicates /Z /FREE /CI /DT=dupsWv Wv
	else 
		FindDuplicates /Z /FREE /DN=dupsWv Wv
	endif
	
	if (numpnts(dupsWv) > 0)
		sprintf error,"contains %d duplicate values",numpnts(dupsWv)
		return [0,error]
	endif
	
	// check for string datatypes
	if (type == 0) 
		return [1,""]
	endif
	
	// get some higher moment information 
	WaveStats /Q Wv
	
	// check for NaNs 
	if (V_numNans)
		sprintf error,"contains %d NaN values",V_numNans
		return [0,error]
	endif
	
	// check for infs
	if (V_numInfs)
		sprintf error,"contains %d ±Inf values",V_numInfs
		return [0,error]
	endif

	// check if a straight line with a gradient (monotonic)
	if ((V_skew == 0) && (V_min != V_max))
		return [1,""]
	endif

	// check indentical values - straight line with no gradient
	if ((V_skew == 0) && (V_min == V_max))
		sprintf error,"contains consecutively identical values (%g)",V_avg
		return [0,error]
	endif
	
	// check for monotonicity in non-linear co-ordinates if necessary (inefficient)
	for (i=2;i<j;i+=1)
		if (abs(Wv[i] - k) <= abs(Wv[i-1] - k))
			sprintf error,"breaks monotonicity at index %d",i
			return [0,error]
		endif
	endfor
	
	return [1,""] // some other case not covered by above e.g. log
End
//___________________________________________________________________________________________________	
// check to see if variable name is CF compliant - luckily also Igor compiant
Static Function isCFName(String WvStr)

	return (CmpStr(CleanUpName(WvStr,0),WvStr[0,strMaxLen-1])==0)
End
//___________________________________________________________________________________________________	
// get the size of a dimension, unlimited dimensions return -1
Static Function getMaxD(String DimPathStr,int dim,double fileID)

	int dummy
	STRUCT HDF5DataInfo di	
	InitHDF5DataInfo(di)
	dummy = HDF5DatasetInfo(fileID, dimPathStr, 0, di)

	return di.maxdims[dim]
End
//___________________________________________________________________________________________________	
// check to see if any previously written dimensions are unlimited
Static Function UDimExists(double fileID)
	
	// get a list of all the variables in the file
	HDF5ListGroup /R fileID, "/"
	
	int i,j = ItemsInList(S_HDF5ListGroup) ; String varStr  
	for (i=0;i<j;i+=1)
		varStr = StringFromList(i,S_HDF5ListGroup)
		HDF5ListAttributes /Type=2 fileID, varStr
		if (GrepString(S_HDF5ListAttributes,"_Netcdf4Dimid"))
			 if(getMaxD(varStr,0,fileID)== -1)
			 	return 1
			 endif
		endif
	endfor
	
	return 0
End
//___________________________________________________________________________________________________	
// the _nc3_strict attribute is in a scalar dataspace
Static Function setClassic(double fileID)
	
	Make /O /FREE /N=1 /Y=(32) attWv = 1
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /O /A="_nc3_strict" attWv, fileID,"/"
	
	return V_flag
End
//___________________________________________________________________________________________________	
// set the NetCDF4DimID attribute in a scalar dataspace 
Static Function setDimID(String path,int id,double fileID)
	
	Make /O /FREE /N=1 /Y=(32) attWv = id
	HDF5SaveData /Z /IGOR=0 /OPTS=1 /O /A="_Netcdf4Dimid" attWv, fileID, path
	
	return V_flag
End
//___________________________________________________________________________________________________	
// function to get a list of netCDF/HDF5 dimensions from DIMENSION_LIST attribute
Static Function/S getHDF5Dimensions(Variable groupID, String WvStr)

	Variable i ; String rtnStr=""
	
	HDF5ListAttributes /Type=2 /Z groupID, WvStr
	
	// variabe has dimensions attached
	if (GrepString(S_HDF5ListAttributes,"DIMENSION_LIST"))
	
		// load the dimension list attribute 
		HDF5LoadData /O /A="DIMENSION_LIST" /TYPE=2 /N=temp_DIMENSION_LIST /Q /Z groupID, WvStr
		
		// get a list of all possible dimension list waves
		String dimWvLst = WaveList("temp_DIMENSION_LIST*",";","")
		if (V_flag==0)	
			for (i=0;i<ItemsInList(dimWvLst);i+=1)
				Wave/T temp_DimLstWv = $StringFRomLIst(i,dimWvLst)
				String dimStr = temp_DimLstWV[0] 
				rtnStr+=dimstr[2,inf]+";"
				KillWaves temp_DimLstWv
			endfor
		endif
	endif
	
	// is a dimension
	HDF5DimensionScale /Z dataset={groupID,WvStr}, isScale
	if (V_Value)
		rtnStr=WvStr
	endif
	
	return rtnStr
End	
//___________________________________________________________________________________________________	
// return the datatype of a variable in a HDF5 file
Static Function typeOnDisk(String path,double fileID)
	
	STRUCT HDF5DataInfo di	
		
	InitHDF5DataInfo(di)	
	int dummy = HDF5DatasetInfo(fileID, path, 0, di)
	int type = h5DataType2WaveType(di.datatype_str)
	
	// check for variable length strings to differentiate CHAR
	type = (type == 0) ? type + (!di.datatype_is_vlen) : type 

	return type
End
//___________________________________________________________________________________________________	
// return the compression of a variable in a HDF5 file
Static Function [int chunk,int deflate,int shuffle,Wave chunksizes] compOnDisk(String path,double fileID)
	
	STRUCT HDF5DataInfo di	
		
	InitHDF5DataInfo(di)	
	int dummy = HDF5DatasetInfo(fileID, path, 4, di)
	
	chunk = (di.layout == 2) ? 1 : 0
	deflate = (strlen(di.filters) == 0) ? 0 : NumberByKey("GZIP",di.filters) 
	shuffle = (strlen(di.filters) == 0) ? 0 : GrepString(di.filters,"SHUFFLE")
	Make /O /FREE /N=(di.ndims) chunksizes = di.chunks[p]
	
	return [chunk,deflate,shuffle,chunksizes]
End
//___________________________________________________________________________________________________	
// get the number of bytes per point for each wavetype
Static Function byteSize(int type)

	int size
	
	switch(type)
		case 0: 	// string
			size=4	// default if /T=num is not specified
			break
		case 2:		// float
			size=4
			break
		case 4:		// double
			size=8
			break
		case 8:		// byte
			size=1
			break
		case 16:	// short
			size=2
			break
		case 32:	// int,long
			size=4
			break
		case 72:	// ubyte
			size=1
			break
		case 80:	// ushort  
			size=2 
			break
		case 96: 	// uint,ulong 
			size=4
			break
		case 128: 	// int64,llong
			size=8
			break
		case 192:  	// uint64,ullong
			size=8
			break
	endswitch
	
	return size
End
//___________________________________________________________________________________________________	
// get a default chunksize for each dimension 
// based on nc4_find_default_chunksizes2 credit: Ed Hartnett, Dennis Heimbigner
Function/WAVE getChunkSizes(Wave Wv,Wave dimsWv)

	int type=WaveType(Wv),i,j=WaveDims(Wv)
	int byte,total,num_chunks,num_unlim
	double extra,suggested_size
	
	// get the number of bytes each point occupies
	if (type == 0)
		byte = getMaxStrLen(Wv)
	else 
		byte = byteSize(type)
	endif
	
	// make sure the chunk size hasn't been set ridiculously low
	int maxChunk = (chunkMaxBytes < byte) ? byte : chunkMaxBytes
	
	// check the unlimited chunk size for validity
	int unlimChunk = (unlimChunkSize == 0) ? default1DunlimChunk : unlimChunkSize 

	// make a wave for holding the chunksizes
	Make /O /FREE /N=4 chunkWv = NaN
	
	// set chunks in unimited dimensions to the default size (probably 1)
	for (i=0;i<j;i+=1)
		if (dimsWv[i] == -1)
			num_unlim+=1
			chunkWv[i]=unlimChunk
		endif
	endfor

	// 1D unlimited variable 
	if ((j == 1) && (sum(dimsWv) == -1))
		if (maxChunk/byte <= 1)
			suggested_size = 1
		elseif (maxChunk/byte > default1DunlimChunk)
			suggested_size = default1DunlimChunk
		else 
			suggested_size = maxChunk
		endif
		chunkWv[0] = ceil(suggested_size/byte)
	endif
	
	// all dimensions unlimited
	if ((j > 1) && (sum(dimsWv) == -j))
		suggested_size = ceil((maxChunk/byte) ^ (1/j))
		for (i=0;i<j;i+=1)
			chunkWv[i] = suggested_size ? suggested_size : unlimChunk
		endfor
	endif
	
	// pick a chunk size finite dimensions
	for (i=0;i<j;i+=1)
		if (numtype(chunkWv[i])>0)
			suggested_size = ceil((maxChunk/byte) ^ (1/(j - num_unlim)))
			if (suggested_size > DimSize(Wv,i))
				suggested_size = DimSize(Wv,i)
			endif
			chunkWv[i] = suggested_size ? suggested_size : unlimChunk
		endif
	endfor
		
	// if total chunk size if bigger than maxChunk
	// keep shrinking the biggest chunk until it fits 		
	do
		total = checkChunksize(j,chunkWv)
		if (total > maxChunk)
			// find the biggest dimension and half it
			WaveStats /Q /M=1 chunkWv
			chunkWv[V_maxLoc]= ceil(chunkWv[V_maxLoc]/2)
		endif
	while (total > maxChunk)
	
	// if overhanging chunks are not allowed, shrink them to size
	if (!overhang)
		for (i=0;i<j;i+=1)
			num_chunks = ceil(Dimsize(Wv,i) / chunkWv[i])
			if(num_chunks > 0) 
				extra = (num_chunks * chunkWv[i]) - Dimsize(Wv,i)
				chunkWv[i] -= floor(extra / num_chunks)
			endif
		endfor
	endif
	
	// set NaN (unused dimensions) to zero
	chunkWv = (numtype(chunkWv[p])>0) ? 0 : chunkWv[p]
	
	return ChunkWv
End	
//___________________________________________________________________________________________________	
// get the total size of a chunk
Static Function checkChunksize(int dims,Wave chunkWv)
	
	int i ;	uint64 total = 1
	
	for (i=0;i<dims;i++)
		total *= chunkWv[i]
	endfor

	return total
end	
//___________________________________________________________________________________________________	
// QLA error handler 
Static Function getError(String funcStr,String errorStr,int errorType,double fileID)

	String errorLst="Information;Warning;Error;"
	
	// if error==0 stay silent
	
	// print the error in the history area
	if ((error > 0) && (error < 4) && (errorType > 0))
		Print "QLA "+StringFromList(errorType-1,errorLst)+": "+errorStr
	endif
	
	// print successful functions
	if ((error > 1) &&  (error < 4) && (errorType == 0))
		Print "QLA: "+errorStr
	endif
	
	// abort on error
	if ((error > 3) && (errorType == 3))
		if (fileStatus(fileID))
			HDF5CloseFile /Q /Z fileID // close the file if it's open
		endif
		DoAlert /T="QLA netCDF-4 library" 0, "The function: "+funcStr+"\r\rReturned the following critical error:\r"+errorStr
		Abort
	endif
	
	return 0
end
//___________________________________________________________________________________________________



//************* prototype functions ******************


//___________________________________________________________________________________________________
// get maximum string length in a text wave
Static Function getMaxStrLen(Wave/T wv)

	double maxlen,len,i
	do 
		len = strlen(wv[i])
		maxLen = (len > maxLen) ? len : maxLen
		i++
	while (i < numpnts(wv))
	
	return maxLen
End
//___________________________________________________________________________________________________	
// convert a text wave into an array of char
Static Function/WAVE text2char(Wave/T textWv)
	
	double n = numpnts(textWv),maxLen = getMaxStrLen(textWv)
	
	Make /O /FREE /T=1 /N=(n,maxLen) charWv = PadString(textWv[p],maxLen,padValue)[q]

	return charWV
End
//___________________________________________________________________________________________________
// get the variable ID from the path
Static Function path2varID(String path,double fileID)
	
	double varID
	
	return varID
End
//___________________________________________________________________________________________________
// get the path from the variable ID
Static Function/S varID2path(double varID,double fileID) 

	String path
	
	return path
End
//___________________________________________________________________________________________________	
// open a netCDF-4 file
Function QLA_open(String pathStr,int omode)

	double fileID ; String errorStr
	if (omode) // write
		HDF5OpenFile /Z fileID as pathStr
	else	// read only
		HDF5OpenFile /Z /R fileID as pathStr
	endif
	
	// handle errors from HDF5OpenFile
	if (V_flag != 0)
		sprintf errorStr,"Could not open the file requested"  
		getError("QLA_open",errorStr,3,fileID)
	endif
	
	return fileID 
End
//___________________________________________________________________________________________________	
// get information about datasets or attributes in a file
Function/S QLA_inq(String path,int type,double fileID)

	String rtnStr="",errorStr

	if (cmpstr(path[strlen(path)-1],"/") == 0)
		HDF5ListGroup /R=1 /Z /F /CONT=1 /ORDR=1 /TYPE=(2^type) fileID, path
		if (V_flag == 0)
			rtnStr = S_HDF5ListGroup
		endif
	else
		HDF5ListAttributes /Z /TYPE=(2^type) fileID, path
		if (V_flag == 0)
			rtnStr = RemoveFromList(HDF5markup+NCmarkup,S_HDF5ListAttributes)
		endif
	endif
	
	if (V_flag != 0)
		sprintf errorStr,"Could not open the file requested"  
		getError("QLA_inq",errorStr,3,fileID)
	endif

	return rtnStr
End
//___________________________________________________________________________________________________
